"""VapourSynth Debanding Functions Collection"""

__version__ = '0.7.2'

__author_name__, __author_email__ = 'Irrational Encoding Wizardry', 'wizards@encode.moe'
__maintainer_name__, __maintainer_email__ = __author_name__, __author_email__

__author__ = f'{__author_name__} <{__author_email__}>'
__maintainer__ = __author__

if __name__ == '__github__':
    print(__version__)
